#!/usr/bin/perl -w

##  illustrate_des_substitution.pl

##  Avi Kak
##  January 15, 2018

##  This is a demonstration of how you can carry out S-boxes based substitution
##  in DES.  The code shown implements the "Substitution with 8 S-boxes" step 
##  that you see inside the dotted Feistel function in Figure 4 of Lecture 3 notes.

##  IMPORTANT:  This demonstration code does NOT include XORing with the round
##              key that must be carried out on the expanded right-half block
##              before it is subject to the S-boxes based substitution step
##              shown here.

use strict;
use Algorithm::BitVector 1.26;

my $expansion_permutation = [31,  0,  1,  2,  3,  4, 
                              3,  4,  5,  6,  7,  8, 
                              7,  8,  9, 10, 11, 12, 
                             11, 12, 13, 14, 15, 16, 
                             15, 16, 17, 18, 19, 20, 
                             19, 20, 21, 22, 23, 24, 
                             23, 24, 25, 26, 27, 28, 
                             27, 28, 29, 30, 31,  0];

my $s_boxes = { map {$_ => undef} 0..7 };


$s_boxes->{0} = [ [14,4,13,1,2,15,11,8,3,10,6,12,5,9,0,7],
                  [0,15,7,4,14,2,13,1,10,6,12,11,9,5,3,8],
                  [4,1,14,8,13,6,2,11,15,12,9,7,3,10,5,0],
                  [15,12,8,2,4,9,1,7,5,11,3,14,10,0,6,13] ];

$s_boxes->{1} = [ [15,1,8,14,6,11,3,4,9,7,2,13,12,0,5,10],
                  [3,13,4,7,15,2,8,14,12,0,1,10,6,9,11,5],
                  [0,14,7,11,10,4,13,1,5,8,12,6,9,3,2,15],
                  [13,8,10,1,3,15,4,2,11,6,7,12,0,5,14,9] ];
 
$s_boxes->{2} = [ [10,0,9,14,6,3,15,5,1,13,12,7,11,4,2,8],
                  [13,7,0,9,3,4,6,10,2,8,5,14,12,11,15,1],
                  [13,6,4,9,8,15,3,0,11,1,2,12,5,10,14,7],
                  [1,10,13,0,6,9,8,7,4,15,14,3,11,5,2,12] ];

$s_boxes->{3} = [ [7,13,14,3,0,6,9,10,1,2,8,5,11,12,4,15],
                  [13,8,11,5,6,15,0,3,4,7,2,12,1,10,14,9],
                  [10,6,9,0,12,11,7,13,15,1,3,14,5,2,8,4],
                  [3,15,0,6,10,1,13,8,9,4,5,11,12,7,2,14] ];

$s_boxes->{4} = [ [2,12,4,1,7,10,11,6,8,5,3,15,13,0,14,9],
                  [14,11,2,12,4,7,13,1,5,0,15,10,3,9,8,6],
                  [4,2,1,11,10,13,7,8,15,9,12,5,6,3,0,14],
                  [11,8,12,7,1,14,2,13,6,15,0,9,10,4,5,3] ];  

$s_boxes->{5} = [ [12,1,10,15,9,2,6,8,0,13,3,4,14,7,5,11],
                  [10,15,4,2,7,12,9,5,6,1,13,14,0,11,3,8],
                  [9,14,15,5,2,8,12,3,7,0,4,10,1,13,11,6],
                  [4,3,2,12,9,5,15,10,11,14,1,7,6,0,8,13] ];

$s_boxes->{6} = [ [4,11,2,14,15,0,8,13,3,12,9,7,5,10,6,1],
                  [13,0,11,7,4,9,1,10,14,3,5,12,2,15,8,6],
                  [1,4,11,13,12,3,7,14,10,15,6,8,0,5,9,2],
                  [6,11,13,8,1,4,10,7,9,5,0,15,14,2,3,12] ];

$s_boxes->{7} = [ [13,2,8,4,6,15,11,1,10,9,3,14,5,0,12,7],
                  [1,15,13,8,10,3,7,4,12,5,6,11,0,14,9,2],
                  [7,11,4,1,9,12,14,2,0,6,10,13,15,3,5,8],
                  [2,1,14,7,4,10,8,13,15,12,9,0,3,5,6,11] ];


# For the purpose of this illustration, let's just make up the right-half of a 
# 64-bit DES bit block:
my $right_half_32bits = Algorithm::BitVector->new( intVal => 800000700, size => 32 );

# Now we need to expand the 32-bit block into 48 bits:
my $right_half_with_expansion_permutation = $right_half_32bits->permute( $expansion_permutation );

print "expanded right_half_32bits: $right_half_with_expansion_permutation\n";

# The following statement takes the 48 bits back down to 32 bits after carrying
# out S-box based substitutions:
my $output = substitute($right_half_with_expansion_permutation);

print "$output\n";

##  This method implements the step "Substitution with 8 S-boxes" step you see inside
##  Feistel Function dotted box in Figure 4 of Lecture 3 notes.
sub substitute {
    my $expanded_half_block = shift;
    my $output = Algorithm::BitVector->new( size => 32 );
    my @segments = map $expanded_half_block->get_slice([$_*6..($_+1)*6]), 0..7;
    foreach my $sindex (0..@segments-1) {
        my $row = 2*int($segments[$sindex]->get_bit(0)) + int($segments[$sindex]->get_bit(-1));
        my $column = int( $segments[$sindex]->get_slice([1..5]) );
        $output->set_slice([$sindex*4..$sindex*4+4], 
           Algorithm::BitVector->new(intVal => $s_boxes->{$sindex}->[$row][$column], size => 4));
    }
    return $output;
}
